@@ -7,4 +7,9 @@ sub test_json :Local {
     $c->res->body($c->req->body_data->{message});
 }
 
+sub test_nested_for :Local {
+    my ($self, $c) = @_;
+    $c->res->body($c->req->body_data->{nested}->{value});
+}
+
 1;
